#
# Maui implementation of resource reservation
# 
# Björn Hagemeier <b.hagemeier@fz-juelich.de>
# created 2011-01-19
#

package ResourceReservation;

require Exporter;
@ISA = qw(Exporter);

@EXPORT_OK = qw(execute_reservation make_reservation cancel_reservation query_reservation);

use Reporting qw(initial_report failed_report debug_report ok_report);
use Submit qw(submit);

use Data::UUID;


#use strict;

# Change this to your actual MAUI bin directory
my $maui_bin_dir = "/usr/local/maui/bin";


BEGIN {
        initial_report("Resource reservation Maui implementation");
     }


#
#make a reservation
#
# arg1: NJS command
#
#
# The NJS command includes the resources as used by the Submit command,
# and the requested starttime in ISO 8601 format
#
# Returns: 
#
#  TSI_OK
#  reservationID
#
#
# Example:
#
#
#   #TSI_MAKE_RESERVATION
#   #TSI_STARTTIME 2007-11-08T09:46:58+0100
#   #TSI_TIME 3600
#    # etc
#

sub make_reservation {

      my $njs_command = shift;
      #failed_report("Resource reservation NOT SUPPORTED by this TSI.");
      print $njs_command;
 
      my $start_time;
      #extract start time (see above for format)
      if($njs_command =~ m/#TSI_STARTTIME (\S+)/) {
	 $start_time = format_starttime($1);
      }

      #DURATION [[[DD:]HH:]MM:]SS  INFINITY 	duration of the reservation (not needed if ENDTIME is specified)
      # #TSI_TIME can be used directly, as it is formatted as seconds only
      my $duration;
      if($njs_command =~ m/#TSI_TIME (\S+)/) {
	  $duration = $1;
      }

      # #TSI_MEMORY already describes MB, which is expected by Maui
      my $memory;
      if($njs_command =~ m/#TSI_MEMORY (\S+)/) {
	  $memory = $1;
      }
      # maui #MEM=<INTEGER> 
      my %resources;
      $resources{"MEM"} = $memory;

      # #TSI_PROCESSORS_PER_NODE
      my $procs;
      if($njs_command =~ m/#TSI_PROCESSORS_PER_NODE (\S+)/) {
	  $procs = $1;
      }
      # maui #PROCS=<INTEGER>

      # #TSI_TOTAL_PROCESSORS
      my $tasks;
      if($njs_command =~ m/#TSI_TOTAL_PROCESSORS (\S+)/) {
	  $tasks = $1;
	  $procs = 1;
	  # total_processors := tasks * procs
      }

      # #TSI_PROCESSORS
      # exists only for backward compatibility, is always the same as #TSI_PROCESSORS_PER_NODE
      if(!defined $procs && ($njs_command =~ m/#TSI_PROCESSORS (\S+)/)) {
	  $procs = $1;
      }
      $resources{"PROCS"} = $procs;

      # #TSI_HOST_NAME
      # this may not be desirable
      my $host;
      if($njs_command =~ m/#TSI_HOST_NAME (\S+)/) {	  
	  $host = $1;
	  if($host == "none") {
	      undef $host;
	  }
      }

      # #TSI_QUEUE
      my $queue;
      if($njs_command =~ m/#TSI_QUEUE (\S+)/) {
	  $queue = $1;
	  if($queue == "NONE") {
	      undef $queue;
	  }
      }

      # #TSI_SSR

      # #TSI_NODES are TASKS when it comes to Maui reservation
      # e.g. TASKS==5
      my $tasks;
      if($njs_command =~ m/#TSI_NODES (\S+)/) {
	  $tasks = $1;
	  if($tasks == "NONE") {
	      $tasks = 1;
	  }
      }

      my $starttime;
      if($njs_command =~ m/#TSI_STARTTIME (\S+)/) {
	  $starttime = $1;
      }

      my $user;
      my $project;
      if($njs_command =~ m/#TSI_IDENTITY (\S+) (\S+)/) {
	  $user = $1;
	  $project = $2;
      }

      foreach my $resource (keys %resources) {
	  push(my @resource_strings, "$resource=$resources[$resource]")
      }
      my $resource_specification = join(":", @resource_strings);

      my $reservation_command = "$maui_bin_dir/setres ";
      $reservation_command .= "-s $starttime ";
      $reservation_command .= "-d $duration ";
      if(defined $user) {
	  $reservation_command .= "-u $user ";
      }
      if(defined $project && !($project eq "NONE")) {
	  $reservation_command .= "-g $project ";
      }

      # tie my $# uuid, 'OSSP::uuid::tie', "v1";
      # $uuid = ["v1"];
      # untie $uuid;
      my $uuidg = new Data::UUID;
      my $uuid = $uuidg->create();
      my $uuid_str = $uuidg->to_string($uuid);
      $reservation_command .= "-n $uuid_str ";

      $reservation_command .= $resource_specification." ";
      $reservation_command .= "TASKS==$tasks";

      #debug_report("Reservation command: '$reservation_command'");
      print "Reservation command: '$reservation_command'\n";
      my $reservation_response = `$reservation_command`;
      my @reservation_output_array = split("\n", $reservation_response);
      my @error_messages = grep /(^ERROR)/, @reservation_output_array;
      if(@error_messages > 0) {
	  failed_report("Couldn't create reservation:\n".join("\n", @error_messages));
      }

      #find the reservation id in the output
      my @id_array = split("'", $reservation_response);
      my $id_length = @id_array;
      my $reservation_id;
      if($id_length == 3) {
	  $reservation_id = $id_array[1];
      }

      print "Reservation id: <$reservation_id>\n";
      print main::CMD_SOCK "$reservation_id\n";
      ok_report("Reservation id: <$reservation_id>");
      
}

#
#cancel a reservation:
#
#arg1: NJS command
#
# The NJS command includes the reservation ID
#
# Example: 
#
# #TSI_CANCEL_RESERVATION
# #TSI_RESERVATION_REFERENCE job.1234
#

sub cancel_reservation {

      my $njs_command = shift;
      my $reservation_id;
      if($njs_command =~ m/#TSI_RESERVATION_REFERENCE (\S+)/) {
	 $reservation_id = $1;
      }

      my $release_response = `$maui_bin_dir/releaseres $reservation_id`;
      my @release_output_lines = split("\n", $release_response);
      my @error_messages = grep /(^ERROR)/, @release_output_lines;
      if(@error_messages > 0) {
	  failed_report("Reservation could not be cancelled:\n".join("\n", @error_messages));
      } else {
	  ok_report("Cancelled reservation: <$reservation_id>");
      }
}

#
#query a reservation: not supported in this dummy module!
#
#arg1: NJS command
#
# The NJS command includes the reservation ID
#
# Example: 
#
# #TSI_QUERY_RESERVATION
# #TSI_RESERVATION_REFERENCE job.1234
#

#
# Result:
# The TSI replies with TSI_OK and a line containing the reservation status and the start time.

# Status is one of UNKNOWN, INVALID, WAITING, READY, ACTIVE, FINISHED or OTHER
# Start time is in the ISO 8601 format used in make_reservation() (see above)
#

sub query_reservation {

      my $njs_command = shift;
      my $reservation_id;
      if($njs_command =~ m/#TSI_RESERVATION_REFERENCE (\S+)/) {
	 $reservation_id = $1;
      }
      debug_report("Query reservation: <$reservation_id>");      
      failed_report("Resource reservation NOT SUPPORTED by this TSI.");
}


#
#Executes (claims) a reservation request
#
#arg1: reservation id
#arg2: NJS command
#
sub execute_reservation {

      my $res_id = shift;
      my $njs_command = shift;

      debug_report("Claim reservation with id <$res_id>");
      
      #simply submit it...
      submit($njs_command);
      
}


sub format_starttime {
# src fmt: #TSI_STARTTIME 2007-11-08T09:46:58+0100
# dst fmt: [HH[:MM[:SS]]][_MO[/DD[/YY]]] or +[[[DD:]HH:]MM:]SS
    my $orig_starttime = shift;
    
    my @date_time = split("T", $orig_starttime);
    my $date = $date_time[0];
    my $time = $date_time[1];

    my @date_components = split("-", $date);
    # this ignores the time zone
    ($plain_time, $time_zone) = split('\+', $time);
    my @time_components = split(":", $plain_time);
    my %date_time = { "year"   => $date_components[0],
		      "month"  => $date_components[1],
		      "day"    => $date_components[2],
		      "hour"   => $time_components[0],
		      "minute" => $time_components[1],
		      "second" => $time_components[2] };
	
    my $date_time_str = join("_", $plain_time, join("/",$date_time["month"],$date_time["day"],$date_time["year"]));

    return $date_time_str;
    
}
